package controller;

import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.Random;

import javax.swing.event.MouseInputListener;

import model.OceanInterface;
import model.OceanObject;
import view.OceanGUIInterface;

/**
 * The listener interface for receiving mouse events. The class that is
 * interested in processing a mouse event implements this interface, and the
 * object created with that class is registered with a component using the
 * component's <code>addMouseListener<code> method. When
 * the mouse event occurs, that object's appropriate
 * method is invoked.
 * 
 * @see MouseEvent
 */
public class MouseListener extends Listener implements MouseWheelListener,
MouseInputListener {

	// _____________________________________________________________________________________

	/**
	 * Instantiates a new mouse listener.
	 * 
	 * @param model the OceanInterface that describes the ocean we work with
	 * @param view the OceanGUIInterface that describes the GUI we work with
	 */
	public MouseListener(OceanInterface model, OceanGUIInterface view) {

		super(model, view);
	}

	// _____________________________________________________________________________________

	/**
	 * Automatically put the coordinates of the mouse click into the x and y
	 * position input fields and click the add button.
	 * 
	 * @param cP the clicked position
	 */
	private void clickAddAndSetPosition(Point cP) {
		getMyOceanGUI().getButtonPanel().getXValueField()
		.setText(Integer.toString(cP.x));
		getMyOceanGUI().getButtonPanel().getYValueField()
		.setText(Integer.toString(cP.y));

		getMyOceanGUI().getButtonPanel().getAddButton().doClick();
	}

	/**
	 * Creates the oceanobject's name.
	 * 
	 * @param type An OceanObject type like: Shark, Fish
	 */
	private void createOcenObjectName(String type) {
		switch (type) {
		case ("Fish"):

			getMyOceanGUI().getButtonPanel().getNameField().setText("Fishy");
		break;

		case ("Shark"):

			getMyOceanGUI().getButtonPanel().getNameField().setText("Sharky");
		break;

		case ("Stone"):

			getMyOceanGUI().getButtonPanel().getNameField().setText("Stony");
		break;

		case ("Plant"):

			getMyOceanGUI().getButtonPanel().getNameField().setText("Planty");
		break;

		case ("Bubble"):

			getMyOceanGUI().getButtonPanel().getNameField().setText("Bubbly");
		break;

		case ("Coke Can"):

			getMyOceanGUI().getButtonPanel().getNameField().setText("CokeCany");
		break;

		}
	}

	/**
	 * Gets the object from click position.
	 * 
	 * @param cP clicked Position
	 * @return the clicked object, or null.
	 */
	private OceanObject getObjectFromClickPosition(Point cP) {

		for (OceanObject obj : getMyOcean().getOceanObjects()) {

			if (

					(((int) cP.getX() >= obj.getPosition()[0]) && ((int) cP.getX() <= (obj
							.getPosition()[0] + obj.getIcon().getIconWidth())))

							&& (((int) cP.getY() >= obj.getPosition()[1]) && ((int) cP
									.getY() <= (obj.getPosition()[1] + obj.getIcon()
											.getIconHeight())))

					)
				return obj;
		}
		return null;
	}

	/**
	 * This method gets an MouseEvent e an executes the specific code.
	 * 
	 * @param e the generated mouse event
	 */
	@Override
	public void mouseClicked(MouseEvent e) {

		Point cP = e.getPoint();
		int mouseButton = e.getButton();

		if (mouseButton == 3) { // right button

			clickAddAndSetPosition(cP);

		} else if (mouseButton == 2) { // middle button/mouse wheel

			setAllFieldsRandom(e);
		}

		if (mouseButton == 1) { // left button

			tagAndSetFields(e, cP);

		}
	}

	@Override
	public void mouseDragged(MouseEvent arg0) {
	}

	@Override
	public void mouseEntered(MouseEvent arg0) {
	}

	@Override
	public void mouseExited(MouseEvent arg0) {
	}

	@Override
	public void mouseMoved(MouseEvent arg0) {
	}

	@Override
	public void mousePressed(MouseEvent arg0) {
	}

	@Override
	public void mouseReleased(MouseEvent arg0) {
	}

	/**
	 * 
	 * This method speeds up or slows down the ocean's movement.
	 * 
	 * @param e the MouseWheelEvent that occurs when the wheel is turned
	 * 
	 */
	@Override
	public void mouseWheelMoved(MouseWheelEvent e) {

		int lowerThreshold = 1000; // ms
		int upperThreshold = 4; // ms
		int currentDelay = getMyOcean().getAnimationDelay();
		int granularity = 0;

		if (currentDelay < 20) {
			granularity = 2;
		} else {
			granularity = currentDelay / 8;
		}

		switch (e.getWheelRotation()) {
		case 1:
			currentDelay += granularity;
			break;
		case -1:
			currentDelay -= granularity;
			break;
		}

		if (currentDelay >= lowerThreshold) {
			currentDelay = lowerThreshold;
		} else if (currentDelay <= upperThreshold) {
			currentDelay = upperThreshold;
		}
		getMyOcean().setAnimationDelay(currentDelay);
	}

	/**
	 * Set all fields (name, maximumAge, x/y-value) random.
	 * 
	 * @param e the MouseEvent
	 */
	private void setAllFieldsRandom(MouseEvent e) {
		Random rand = new Random();
		String type = getMyOceanGUI().getButtonPanel()
				.getOceanObjectTypesString();

		getMyOceanGUI()
		.getButtonPanel()
		.getXValueField()
		.setText(
				Integer.toString(rand
						.nextInt(getMyOcean().getWidth() - 90)));
		getMyOceanGUI()
		.getButtonPanel()
		.getYValueField()
		.setText(
				Integer.toString(rand
						.nextInt(getMyOcean().getDepth() - 90)));

		String maxAge = new String();

		if ((e.getModifiers() & ActionEvent.SHIFT_MASK) != 0) {
			maxAge = Integer.toString(25000);
		} else {
			maxAge = Integer.toString(rand.nextInt(1000));
		}
		getMyOceanGUI().getButtonPanel().getMaximumAgeField().setText(maxAge);
		createOcenObjectName(type);
	}

	/**
	 * Tag the object and set the x/y-value-fields (left-Button)
	 * 
	 * @param e MouseEvent
	 * @param cP clicked Position.
	 */
	private void tagAndSetFields(MouseEvent e, Point cP) {

		getMyOceanGUI().getButtonPanel().getXValueField()
		.setText(Integer.toString(cP.x));
		getMyOceanGUI().getButtonPanel().getYValueField()
		.setText(Integer.toString(cP.y));

		OceanObject obj = getObjectFromClickPosition(cP);
		if (obj != null) {

			obj.setTagged(!obj.getTagged());
			updateGUI();
		}
	}
}
