package model;

import static model.OceanObject.KEINE_KOLLISION;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.LinkedList;

import stuff.Pair;

/**
 * The Class Ocean.
 */
public class Ocean implements OceanInterface, Serializable {

	/**
	 * The Class EmptyListException.
	 *
	 * Instances of this class are thrown whenever the user clicks
	 * the remove button without having selected objects to remove.
	 */
	public class EmptyListException extends Exception {

		/** The Constant serialVersionUID. */
		private static final long serialVersionUID = 7312146429306850592L;
	}

	/**
	 * The Class PositionOccupiedException.
	 *
	 * Instancs of this class are thrown whenever the user tries
	 * to add an object to the ocean at a position where another object is.
	 */
	public static class PositionOccupiedException extends Exception {

		/** The Constant serialVersionUID. */
		private static final long serialVersionUID = 7312146429306850592L;
	}

	/**
	 * The Class PositionOutOfBoundsException.
	 *
	 * Instancs of this class are thrown whenever the user tries
	 * to add an object to the ocean at a position too close to the ocean
	 * boundaries or entirely outside the ocean.
	 */
	public static class PositionOutOfBoundsException extends Exception {

		/** The Constant serialVersionUID. */
		private static final long serialVersionUID = 7312146429306850592L;
	}

	/** The Constant BUBBLE. */
	public static final int BUBBLE = 3;

	/** The Constant COKE_CAN. */
	public static final int COKE_CAN = 6;

	/** The Constant FISH. */
	public static final int FISH = 0;

	/** The Ocean object id. */
	private static int OceanObjectID = 0;

	/** The Constant oceanObjectTypes. */
	public static final LinkedList<OceanObject> oceanObjectTypes = new LinkedList<OceanObject>();

	/** The Constant PLANT. */
	public static final int PLANT = 4;

	/** The Constant serialVersionUID. */
	private static final long serialVersionUID = 1L;

	/** The Constant SHARK. */
	public static final int SHARK = 1;

	/** The Constant STONE. */
	public static final int STONE = 2;

	/** The ocean's age. */
	private int age = 0;

	/** The delay between to steps of the animation. */
	private int animationDelay = 30;

	/** The ocean's depth. */
	private int depth;

	/** Says whether or not the ocean is running. */
	private boolean isOceanRunning = false;

	/** The ocean's movement thread. */
	private MovementThread mt;

	/** Holds the objects that are added to the ocean after all the collisions have been dealt with. */
	private LinkedList<OceanObject> newObjects = new LinkedList<OceanObject>();

	/** The ocean's object list. */
	private LinkedList<OceanObject> oceanObjects = new LinkedList<OceanObject>();

	/** The ocean's width. */
	private int width;

	/**
	 * Instantiates a new ocean.
	 *
	 * @param width the ocean's width
	 * @param depth the ocean's depth
	 */
	public Ocean(int width, int depth) {

		this(width, depth, new LinkedList<OceanObject>());

	}

	/**
	 * Instantiates a new ocean.
	 *
	 * @param width the ocean's width
	 * @param depth the ocean's depth
	 * @param oceanObjects the ocean objects
	 */
	public Ocean(int width, int depth, LinkedList<OceanObject> oceanObjects) {

		setWidth(width);
		setDepth(depth);
		setOceanObjects(oceanObjects);
		mt = new MovementThread(this, 42);
		mt.start();
	}
	/**
	 * Gets the ocean's age.
	 *
	 * @return age the age
	 */
	public int getAge() {
		return age;
	}

	/**
	 * Gets the ocean's animation delay.
	 *
	 * @return animationDelay the animationDelay
	 */
	@Override
	public int getAnimationDelay() {
		return animationDelay;
	}

	/**
	 * Gets the ocean's depth.
	 *
	 * @return depth the ocean's depth
	 * 
	 */
	@Override
	public int getDepth() {

		return this.depth;

	}

	/**
	 *
	 * Gets the movementThread.
	 *
	 * @return mt the ocean's movementThread
	 */
	@Override
	public MovementThread getMovementThread() {
		return mt;
	}

	/**
	 * Gets the new objects.
	 *
	 * @return newObjects the newObjects
	 */
	public LinkedList<OceanObject> getNewObjects() {
		return newObjects;
	}

	/**
	 * Gets the number of ocean objects.
	 *
	 * @return the number of ocean objects
	 */
	public int getNumberOfOceanObjects() {
		return oceanObjects.size();
	}

	/**
	 * Returns the number of the special objects in the ocean.
	 *
	 * @param type the object's type
	 * @return the number of special objects
	 */
	@Override
	public int getNumberOfOceanObjectsByType(int type) {
		int result = 0;
		synchronized (getOceanObjects()) {
			for (int k = getOceanObjects().size() - 1; k >= 0; k--) {
				OceanObject obj = getOceanObjects().get(k);
				if (obj.getType() == type) {
					result++;
				}
			}
		}
		return result;
	}

	/**
	 * Gets the number of ocean objects without bubbles.
	 *
	 * @return the number of ocean objects without bubbles
	 */
	public int getNumberOfOceanObjectswithoutBubbles() {
		int population = 0;

		for (int k = 0; k < getNumberOfOceanObjects(); k++) {

			if (getOceanObjects().get(k).getType() != BUBBLE) {
				population += 1;
			}
		}
		return population;
	}


	/**
	 * Returns the list of all oceanObjects.
	 *
	 * @return the list of ocean objects
	 */
	@Override
	public LinkedList<OceanObject> getOceanObjects() {
		return oceanObjects;
	}


	/**
	 *
	 * The method is a helper for the statusPanel.
	 *
	 * @return Info-Pair of the ocean's age and the oceanobjectsList size
	 */
	@Override
	public Pair<Integer, Integer> getStatusPanelInfo() {
		return new Pair<Integer, Integer>(getAge(), getNumberOfOceanObjectswithoutBubbles());
	}

	/**
	 * Gets the width.
	 *
	 * @return the ocean's width
	 *
	 */
	@Override
	public int getWidth() {
		return this.width;
	}

	/**
	 * Increase the ocean's age.
	 */
	public void incAge() {
		this.age++;

	}

	/**
	 * Checks if the ocean is running.
	 *
	 * @return whether or not the ocean is running
	 */
	@Override
	public boolean isOceanRunning() {
		return this.isOceanRunning;
	}


	/**
	 * Sets the ocean's age.
	 *
	 * @param age the new age
	 */
	public void setAge(int age) {
		this.age = age;
	}

	/**
	 * Sets the animation delay.
	 *
	 * @param animationDelay            the animationDelay to set
	 */
	@Override
	public void setAnimationDelay(int animationDelay) {
		this.animationDelay = animationDelay;
	}

	/**
	 * Sets the depth.
	 *
	 * @param depth the new depth
	 * 
	 */
	@Override
	public void setDepth(int depth) {

		this.depth = depth;

	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#setMovementThread(model.MovementThread)
	 */
	@Override
	public void setMovementThread(MovementThread movementThread) {
		this.mt = movementThread;
	}

	/**
	 * Sets the new objects.
	 *
	 * @param newObjects            the newObjects to set
	 */
	public void setNewObjects(LinkedList<OceanObject> newObjects) {
		this.newObjects = newObjects;
	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#setOceanObjects(java.util.LinkedList)
	 */
	@Override
	public void setOceanObjects(LinkedList<OceanObject> oceanObjects) {

		this.oceanObjects = oceanObjects;

	}

	/**
	 * Sets the ocean running.
	 *
	 * @param isRunning the new ocean running
	 */

	public void setOceanRunning(boolean isRunning) {
		this.isOceanRunning = isRunning;
		if (isOceanRunning()) {

			synchronized (getMovementThread().getLock()) {
				getMovementThread().getLock().notify();
			}

			synchronized (getOceanObjects()) {

				for (int k = getOceanObjects().size() - 1; k >= 0; k--) {
					OceanObject obj = getOceanObjects().get(k);
					if (obj.getType() == COKE_CAN) {
						CokeCan c = (CokeCan) obj;
						CokeCanBubbleGeneratorThread t = c.getCcbgt();
						if (t != null) {
							System.out.println("Wecke BubbleThread von Dose "
									+ obj.getOceanObjectName() + " auf");
							synchronized (t.getLock()) {
								t.getLock().notify();
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Sets the width.
	 *
	 * @param width the new width
	 * 
	 */
	@Override
	public void setWidth(int width) {

		this.width = width;

	}


	/**
	 * Adds the newly spawned objects to ocean.
	 *
	 * @param newObjects the new objects
	 * @throws PositionOccupiedException if the position is occupied
	 * @throws PositionOutOfBoundsException if the position is out of bounds
	 */
	private synchronized void addNewlySpawnedObjectsToOcean(
			LinkedList<OceanObject> newObjects) throws Exception {

		synchronized (getOceanObjects()) {

			for (int k = getNewObjects().size() - 1; k >= 0; k--) {
				OceanObject o = getNewObjects().get(k);
				addOceanObject(o);
			}
		}
		getNewObjects().clear();
	}

	/**
	 * Make the oceanObject's ID.
	 *
	 * @return an unused OceanObjectID, as long as there are less than 2<sup>31</sup>-1 objects in the ocean
	 */
	public static int makeOOID() {
		OceanObjectID = OceanObjectID + 1;
		return OceanObjectID;
	}

	/**
	 *
	 * This method adds the given object to the list.
	 *
	 * @param obj the object to be added
	 */
	@Override
	public synchronized void addOceanObject(OceanObject obj)
			throws PositionOccupiedException, PositionOutOfBoundsException {

		try {

			if (positionOccupied(obj))
				throw new PositionOccupiedException();

			if (obj.collidesWithOceanBounds() != KEINE_KOLLISION)
				throw new PositionOutOfBoundsException();

			getOceanObjects().add(obj);
		}

		finally {
			/* syntax error, insert "Finally" to complete block statements */
		}
	}

	/**
	 *
	 * This method checks the ocenObjectList; is the given object in the list?
	 *
	 * @param o
	 * @return whether the object is in the oceanObject list
	 *
	 */
	@Override
	public Boolean contains(OceanObject o) {
		synchronized (getOceanObjects()) {
			return getOceanObjects().indexOf(o) != -1;
		}
	}

	/**
	 * Load a saved ocean.
	 *
	 * @param path the load location
	 * @param name of the saved file
	 *
	 */
	@Override
	public Boolean load(String path, File name) {

		try {

			FileInputStream inStream = new FileInputStream(path + name);
			ObjectInputStream in = new ObjectInputStream(inStream);

			Ocean readOcean = (Ocean) in.readObject();

			int readOceanWidth = readOcean.getWidth();
			int readOceanDepth = readOcean.getDepth();
			LinkedList<OceanObject> readOceanObjects = readOcean
					.getOceanObjects();

			this.width = readOceanWidth;
			this.depth = readOceanDepth;
			this.oceanObjects = readOceanObjects;

			in.close();
			inStream.close();

			makeNewThreadsForCokeCans();

			return true;
		}

		catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * Make new threads for coke cans after they have been loaded.
	 */
	private void makeNewThreadsForCokeCans() {

		synchronized (getOceanObjects()) {

			for (int k = getOceanObjects().size() - 1; k >= 0; k--) {
				OceanObject o = getOceanObjects().get(k);

				if (o.getType() == COKE_CAN) {

					CokeCan c = (CokeCan) o;
					CokeCanBubbleGeneratorThread t = null;

					System.out.println("Dose " + c.getOceanObjectName() + " ("
							+ c.getID() + ") hat den Boden"
							+ (!c.hasReachedBottom() ? " noch nicht" : "")
							+ " erreicht");

					if (c.hasReachedBottom()) {
						t = new CokeCanBubbleGeneratorThread(this, c, o.getID());
						c.setCcbgt(t);
						t.start();
						System.out.println("ccbgt of " + o.getOceanObjectName()
								+ " is " + t.getState());
					}
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#move()
	 */
	@Override
	public synchronized void move() throws Exception {

		LinkedList<Pair<OceanObject, OceanObject>> collisions = new LinkedList<Pair<OceanObject, OceanObject>>();

		synchronized (getOceanObjects()) {

			for (int k = getOceanObjects().size() - 1; k >= 0; k--) {

				OceanObject outer = getOceanObjects().get(k);
				outer.move();

				for (int h = k + 1; h <= (getOceanObjects().size() - 1); h++) {

					OceanObject inner = getOceanObjects().get(h);

					if (outer.movementCollisionTwo(inner)) {

						Pair<OceanObject, OceanObject> newReactionPair = new Pair<OceanObject, OceanObject>(
								outer, inner);

						collisions.add(newReactionPair);
					}
				}
				processStalledCollisions(collisions);
				removeAllDeadObjects();
				addNewlySpawnedObjectsToOcean(getNewObjects());
			}
		}
		incAge();
	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#positionOccupied(model.OceanObject)
	 */
	@Override
	public Boolean positionOccupied(OceanObject o) {
		Boolean result = false;

		if (getOceanObjects().size() == 0)
			return false;

		synchronized (getOceanObjects()) {

			for (int k = getOceanObjects().size() - 1; k >= 0; k--) {
				OceanObject objInOcean = getOceanObjects().get(k);
				if (objInOcean.RectangleIntersect(o)) {
					result = true;
				}
			}
		}
		return result;
	}

	/**
	 * This method deals with all the collisions that have occurred between ocean objects.
	 *
	 * @param collisions the collisions
	 */
	private synchronized void processStalledCollisions(
			LinkedList<Pair<OceanObject, OceanObject>> collisions) {

		for (Pair<OceanObject, OceanObject> p : collisions) {

			p.getFirst().reaktion(p.getSecond());

		}
	}

	/**
	 * Removes all the dead objects from the ocean.
	 */
	public void removeAllDeadObjects() {

		LinkedList<OceanObject> l = getOceanObjects();

		synchronized (l) {

			for (int h = l.size() - 1; h >= 0; h--) {

				OceanObject obj = l.get(h);

				if (obj.getMaximumAge() <= obj.getAge()) {
					removeOceanObject(obj);
				}

			}

		}

	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#removeAllObjects()
	 */
	@Override
	public void removeAllObjects() {

		LinkedList<OceanObject> l = getOceanObjects();

		synchronized (l) {

			for (int h = l.size() - 1; h >= 0; h--) {

				OceanObject obj = l.get(h);

				removeOceanObject(obj);

			}

		}

	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#removeAllTaggedObjects()
	 */
	@Override
	public void removeAllTaggedObjects() throws EmptyListException {

		LinkedList<OceanObject> l = getOceanObjects();

		synchronized (l) {

			int numberOfRemovedObjects = 0;

			for (int h = l.size() - 1; h >= 0; h--) {

				OceanObject obj = l.get(h);

				if (obj.getTagged()) {

					removeOceanObject(obj);
					numberOfRemovedObjects++;
				}

			}

			if (numberOfRemovedObjects == 0)
				throw new EmptyListException();

		}

	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#removeOceanObject(model.OceanObject)
	 */
	@Override
	public void removeOceanObject(OceanObject o) {
		LinkedList<OceanObject> l = getOceanObjects();

		synchronized (l) {
			l.remove(o);
		}
	}

	/**
	 * Removes the threads from the coke cans.
	 */
	private void removeThreadsFromCokeCans() {

		synchronized (getOceanObjects()) {

			for (int k = getOceanObjects().size() - 1; k >= 0; k--) {
				OceanObject o = getOceanObjects().get(k);
				if (o.getType() == COKE_CAN) {
					CokeCan c = (CokeCan) o;
					c.setCcbgt(null);
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#save(java.lang.String, java.io.File)
	 */
	@Override
	public Boolean save(String path, File name) {

		try {

			removeThreadsFromCokeCans();
			this.isOceanRunning = true;
			FileOutputStream outStream = new FileOutputStream(path + name);
			ObjectOutputStream out = new ObjectOutputStream(outStream);

			out.writeObject(this);

			out.close();
			outStream.close();
			this.isOceanRunning = false;
			return true;

		}

		catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#start()
	 */
	@Override
	public void start() {
		setOceanRunning(true);
	}

	/* (non-Javadoc)
	 * @see model.OceanInterface#stop()
	 */
	@Override
	public void stop() {
		setOceanRunning(false);
	}

	/**
	 * Generates a textual representation of the ocean.
	 *
	 * @return The ocean objects in a string
	 */
	@Override
	public String toString() {

		String oceanState = new String();

		oceanState = oceanState + "The ocean has a width of " + getWidth()
				+ " and a depth of " + getDepth()
				+ " units\nand contains the following objects:\n\n";

		if (getOceanObjects().size() == 0) {

			oceanState = oceanState + "[none]\n\n----------\n";

		} else {

			for (OceanObject obj : getOceanObjects()) {
				oceanState = oceanState + obj.toString();
			}

			oceanState = oceanState + "\n----------\n";
		}

		return oceanState;

	}
}
