package model;

import java.io.File;
import java.util.LinkedList;

import model.Ocean.EmptyListException;
import model.Ocean.PositionOccupiedException;
import model.Ocean.PositionOutOfBoundsException;
import stuff.Pair;

public interface OceanInterface {

	/**
	 *    With this method, you add a new object to the
	 *             OceanObjectList (also to the ocean).
	 * 
	 * @param obj
	 * @throws PositionOccupiedException if the position is occupied
	 * @throws PositionOutOfBoundsException if the position is out of bounds
	 * 
	 */
	public void addOceanObject(OceanObject obj)
			throws PositionOccupiedException, PositionOutOfBoundsException;

	/**
	 * 
	 * @param o the ocean object to be added
	 * @return whether or not o is in the ocean object list
	 */
	public Boolean contains(OceanObject o);


	/**
	 * @return the delay between two steps of the animation
	 */
	public int getAnimationDelay();

	/**
	 * @return Get the ocean's depth.
	 */
	public int getDepth();

	/**
	 * Gets the movement thread
	 * 
	 * @return the ocean's movement thread
	 */
	public MovementThread getMovementThread();

	/**
	 * @param type the type of ocean object to count
	 * @return The number of ocean objects of the specified type in the
	 *         OceanObjectList
	 */
	public int getNumberOfOceanObjectsByType(int type);

	/**
	 * 
	 * @return The list of ocean objects.
	 */
	public LinkedList<OceanObject> getOceanObjects();

	/**
	 * @return a pair of integers containing the number of oceanObjects (sans bubbles) and the ocean's age.
	 */
	public Pair<Integer, Integer> getStatusPanelInfo();

	/**
	 * @return the ocean's width
	 */
	public int getWidth();

	/**
	 * 
	 * @return whether or not the ocean is running
	 */
	public boolean isOceanRunning();

	/**
	 * @param path the location of the ocean to load
	 * @param name the ocean's file name
	 * @return whether or not the ocean could be loaded
	 */
	public Boolean load(String path, File name);

	/**
	 * This method moves all the ocean's objects.
	 * 
	 * @throws Exception hands exception received from the
	 * addNewlySpawnedObjects() method to the controller
	 * 
	 */
	public void move() throws Exception;

	/**
	 * Determines whether the position of the object o is occupied
	 * (i.e. whether the object o can be added there).
	 * 
	 * 
	 * @param o an ocean object to be added
	 * @return whether or not the position is occupied
	 */
	public Boolean positionOccupied(OceanObject o);

	/**
	 * This method removes all objects from the OceanObjectList.
	 */
	public void removeAllObjects();

	/**
	 * This method removes all tagged objects from the OceanObjectList.
	 * 
	 * @throws EmptyListException if no objects have been selected for removal.
	 * 
	 */
	public void removeAllTaggedObjects() throws EmptyListException;

	/**
	 * This method removes a specific object from the OceanObjectList.
	 * 
	 * @param o the ocean object to be removed
	 */
	public void removeOceanObject(OceanObject o);

	/**
	 * This method saves an ocean to a file.
	 * 
	 * @param path the location to save the ocean at
	 * @param name the file name for the ocean
	 * @return whether or not the ocean could be saved
	 */
	public Boolean save(String path, File name);

	/**
	 * This method sets the delay between two steps of the animation.
	 * 
	 * @param currentDelay the animation delay to set
	 */
	public void setAnimationDelay(int currentDelay);

	/**
	 * This method sets the ocean's depth.
	 * 
	 * @param depth the depth to set
	 */
	public void setDepth(int depth);

	/**
	 * This method sets the movement thread
	 * 
	 * @param mt the movement thread to set
	 */
	public void setMovementThread(MovementThread mt);

	/**
	 * This method sets the ocean object list.
	 * 
	 * @param oceanObjects the list of all OceanObjects in the ocean.
	 */
	public void setOceanObjects(LinkedList<OceanObject> oceanObjects);

	/**
	 * This method sets the ocean's width.
	 * @param width the oceans width
	 */
	public void setWidth(int width);

	/**
	 * This method starts the movement.
	 */
	public void start();

	/**
	 * This method stops the movement, all objects in the list freeze in the
	 * current position.
	 */
	public void stop();

}
